import ipaddress
import subprocess
from typing import Optional

sans_ipctl_bin = "ipctl"

isg_group = "adm-soseth"


class SansError(Exception):
    pass


class NotAuthenticatedError(SansError):
    """The credentials did not work."""

    pass


class IPNotFoundError(Exception):
    pass


class SubnetNotFoundError(Exception):
    pass


class NotStaticSubnet(Exception):
    pass


class NotDHCPSubnet(Exception):
    pass


class NoFreeIPError(Exception):
    pass


class Subnet:
    def __init__(
        self,
        name: str,
        vlan: str,
        v4net: str,
        v6net_static_: Optional[str],
        v6net_dhcp_: Optional[str],
        offset: int,
        comment: str,
    ):
        self.name = name
        self.vlan = vlan
        self.v4net = ipaddress.IPv4Network(v4net, strict=True)
        if v6net_static_:
            self.v6net_static: Optional[ipaddress.IPv6Network] = ipaddress.IPv6Network(v6net_static_, strict=True)
        else:
            self.v6net_static = None
        if v6net_dhcp_:
            self.v6net_dhcp: Optional[ipaddress.IPv6Network] = ipaddress.IPv6Network(v6net_dhcp_, strict=True)
        else:
            self.v6net_dhcp = None
        self.offset = offset
        self.comment = comment

    def __repr__(self) -> str:
        return """{} ({})
    IPv4: {} 
    IPv6 (Static): {}
    IPv6 (DHCP): {}
    Offset: 0x{:04x}
    Description: {}""".format(
            self.name,
            self.vlan,
            self.v4net,
            self.v6net_static,
            self.v6net_dhcp,
            self.offset,
            self.comment,
        )

    def find_free_ip(self) -> ipaddress.IPv4Address:
        subnet = self.v4net.network_address

        try:
            out = subprocess.check_output([sans_ipctl_bin, "ip", "free", str(subnet)])
        except subprocess.CalledProcessError as e:
            if e.returncode == 10:
                raise NotAuthenticatedError
            raise SansError(e)

        addresses = [x for x in out.decode().split("\n") if x]
        if not addresses:
            raise NoFreeIPError

        return ipaddress.IPv4Address(addresses[0])


class SubnetList:
    def __init__(self, data: list[Subnet]):
        self.netlist: dict[str, Subnet] = {}
        for s in data:
            self.netlist[s.name] = s

    def __repr__(self) -> str:
        r = []
        for k in sorted(self.netlist):
            r.append(repr(self.netlist[k]))
        return "\n".join(r)

    def _find_subnet(self, v4: ipaddress.IPv4Address) -> Subnet:
        for k in self.netlist:
            subnet = self.netlist[k]
            if v4 in subnet.v4net:
                return subnet

        raise SubnetNotFoundError

    def ip4to6(self, ipv4: ipaddress.IPv4Address, dhcp: bool = False) -> ipaddress.IPv6Interface:
        subnet = self._find_subnet(ipv4)

        # find the offset from the start of the subnet.
        # it will then be used to select the corresponding ipv6 address
        idx = list(subnet.v4net).index(ipv4)
        if dhcp:
            if not subnet.v6net_dhcp:
                raise NotDHCPSubnet
            v6addr = subnet.v6net_dhcp[subnet.offset + idx]
            v6_str = "{}/{}".format(v6addr, subnet.v6net_dhcp.prefixlen)
        else:
            if not subnet.v6net_static:
                raise NotStaticSubnet
            v6addr = subnet.v6net_static[subnet.offset + idx]
            v6_str = "{}/{}".format(v6addr, subnet.v6net_static.prefixlen)

        # using an Interface for also storing the subnet prefix
        return ipaddress.IPv6Interface(v6_str)

    def register_addr(
        self, ipv4: ipaddress.IPv4Address, fqdn: str, dry_run: bool = False, no_ipv6: bool = False
    ) -> Optional[ipaddress.IPv6Address]:
        ipv6 = self.ip4to6(ipv4).ip

        if not dry_run:
            try:
                subprocess.check_call(
                    [
                        sans_ipctl_bin,
                        "ip",
                        "add",
                        "",
                        fqdn,
                        "ip:{}".format(ipv4),
                        "reverse:Y",
                        "isgGroup:{}".format(isg_group),
                    ]
                )
            except subprocess.CalledProcessError as e:
                raise SansError(e)

            if not no_ipv6:
                try:
                    subprocess.check_call(
                        [
                            sans_ipctl_bin,
                            "ip",
                            "add",
                            "",
                            fqdn,
                            "ipv6:{}".format(ipv6),
                            "reverse:Y",
                            "isgGroup:{}".format(isg_group),
                        ]
                    )
                except subprocess.CalledProcessError as e:
                    raise SansError(e)

        if no_ipv6:
            return None
        else:
            return ipv6


subnet_data = [
    Subnet(
        "als",
        "ALS::148",
        "129.132.227.64/28",
        None,
        "2001:67c:10ec:49c1:8000::/115",
        0x0100,
        "ALS (sos-als-client-1)",
    ),
    Subnet(
        "hg",
        "HG::147",
        "129.132.36.208/29",
        None,
        "2001:67c:10ec:49c8:8000::/115",
        0x0100,
        "HG (sos-hg-client-1)",
    ),
    Subnet(
        "tur",
        "NW::147",
        "129.132.54.0/26",
        "2001:67c:10ec:49ce::/118",
        "2001:67c:10ec:49ce:8000::/115",
        0x0100,
        "TUR (sos-nw-client-1)",
    ),
    Subnet(
        "tur_private",
        "NW::147",
        "172.31.118.0/25",
        "2001:67c:10ec:49ce::/118",
        "2001:67c:10ec:49ce:8000::/115",
        0x0200,
        "TUR (sos-nw-client-2)",
    ),
    Subnet(
        "tur_spec_1",
        "NW::148",
        "172.31.156.32/28",
        "2001:67c:10ec:49c5::/118",
        None,
        0x0100,
        "TUR (sos-nw-spec-1)",
    ),
    Subnet(
        "tur_spec_2",
        "NW::148",
        "172.31.156.48/28",
        "2001:67c:10ec:49c5::/118",
        None,
        0x0200,
        "TUR (sos-nw-spec-2)",
    ),
    Subnet(
        "vm",
        "DC::2522",
        "192.33.91.0/24",
        "2001:67c:10ec:49c3::/118",
        None,
        0x0100,
        "VM (sos-dc-server-1)",
    ),
    Subnet(
        "mgmt",
        "DC::2523",
        "82.130.108.192/27",
        "2001:67c:10ec:49c4::/118",
        None,
        0x0300,
        "Management (sos-dc-server-2)",
    ),
    Subnet(
        "mgmt_2",
        "DC::2523",
        "82.130.124.80/28",
        "2001:67c:10ec:49c4::/118",
        None,
        0x0100,
        "Management (sos-dc-server-3)",
    ),
    Subnet(
        "ipmi",
        "DC::2999",
        "172.31.0.192/26",
        "2001:67c:10ec:3e00::/118",
        None,
        0x0100,
        "IPMI (sos-lee-dc-server-3)",
    ),
    Subnet(
        "cxsrv",
        "Cx::147",
        "129.132.16.48/28",
        "2001:67c:10ec:49c2::/118",
        None,
        0x0100,
        "CAB/CHN (sos-cx-server-2)",
    ),
    Subnet(
        "cxsrv_2",
        "Cx::147",
        "129.132.74.16/28",
        "2001:67c:10ec:49c2::/118",
        None,
        0x0200,
        "CAB/CHN (sos-cx-server-1)",
    ),
    Subnet(
        "cxsrv_private",
        "Cx::147",
        "192.168.38.64/26",
        "2001:67c:10ec:49c2::/118",
        None,
        0x0300,
        "CAB/CHN (sos-cx-server-3)",
    ),
]

subnets = SubnetList(subnet_data)
