import argparse
import ipaddress
import sys
from typing import Optional

import nettools.sos_infrastructure as sos
from nettools._version import __version__


class CliError(Exception):
    pass


def do_register_addr(args: argparse.Namespace, ipv4_override: Optional[ipaddress.IPv4Address] = None) -> None:
    if not ipv4_override:
        try:
            ipv4 = ipaddress.IPv4Address(args.ipv4)
        except ipaddress.AddressValueError:
            raise CliError("Invalid IPv4 passed as argument")
    else:
        ipv4 = ipv4_override

    try:
        ipv6 = sos.subnets.register_addr(ipv4, args.fqdn, args.dry, args.no_ipv6)

        if ipv6:
            print("{} {} {}".format(args.fqdn, ipv4, ipv6))
        else:
            print("{} {}".format(args.fqdn, ipv4))
    except sos.SubnetNotFoundError:
        raise CliError(f"Failed to find subnet data for address: {ipv4}")
    except sos.IPNotFoundError:
        raise CliError("The entered IP wasn't found")
    except sos.SansError as e:
        raise CliError(f"ERROR: Invoking sans failed: {e}")


def do_register_net(args: argparse.Namespace) -> None:
    subnet = sos.subnets.netlist[args.net_name]
    try:
        ipv4 = subnet.find_free_ip()
    except sos.NoFreeIPError:
        raise CliError("Couldn't find a free IP")

    return do_register_addr(args, ipv4)


def register_host_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("--dry", action="store_true", help="dry run, don't register")
    parser.add_argument("--no-ipv6", action="store_true", help="only register ipv4 address")

    parser.add_argument("--version", action="version", version="%(prog)s " + __version__)

    commands = parser.add_subparsers(dest="command", help="commands")
    commands.required = True

    parser_addr = commands.add_parser("addr", help="register given address")
    parser_addr.add_argument("ipv4", type=str, help="ipv4 address")
    parser_addr.add_argument("fqdn", type=str, help="FQDN")

    parser_net = commands.add_parser("net", help="find free address in network and register")
    parser_net.add_argument("net_name", type=str, help="name of network")
    parser_net.add_argument("fqdn", type=str, help="FQDN")

    commands.add_parser("list", help="list available networks")
    parser.parse_args()

    return parser


def cli() -> int:
    parser = register_host_parser()
    args = parser.parse_args()

    try:
        if args.command == "addr":
            do_register_addr(args)
        elif args.command == "net":
            do_register_net(args)
        elif args.command == "list":
            print(sos.subnets)
    except CliError as e:
        print(f"ERROR: {e}", file=sys.stderr)
        return 1
    except sos.NotAuthenticatedError:
        print("ERROR: sans username or password wrong", file=sys.stderr)
        return 10

    return 0
